unit OSC;

{
Unterschiede zwischen Delphi-COmponent und Active-X Component
- Versionsstring
notes
  problem beim lesen, wenn keine daten vorhanden, dann variante string und conversion problems
  autoopen

  clear buffer, methode und in open
  check memory usage
  simplify oscmsglist
  timestamp
}
{ $Define ActiveX}
interface
uses
  Windows, Messages, SysUtils, Classes, Controls, ExtCtrls,IdBaseComponent, IdAntiFreezeBase, IdAntiFreeze,
  IdComponent, IdUDPBase, IdUDPClient, IdStack, IdUDPServer, IdSocketHandle, Contnrs, Graphics;

const
{$ifdef ActiveX}
  Ver = 'OSC Active-X, Ver 0.92, (c)2006 Frieder Weiss';
{$else}
  Ver = 'OSC Component for Delphi, Ver 0.92, (c)2006 Frieder Weiss';
{$endif}
type
  TOSCmsg = class(TObject)
  public
    bundleID : DWord;
    address : string;
    timestamp: DWORD;
    typetag: string;
    varia: array of OLEVariant;
  end;


  TOSC = class(TCustomPanel)
  private
    FIndicator: integer;
    FProgress: integer;
    FListeningPort: integer;
    FUDPServer: TIdUDPServer;
    FBundleID: integer;
    { Events }
    FOnOSCInput: TNotifyEvent;	{ OSC Input arrived }
    OSCMsgList : TObjectList;
    procedure SetListeningPort(const Value: integer);
    { Private-Deklarationen }
  protected
    { Protected-Deklarationen }
    procedure Paint; override;
    procedure UDPServerUDPRead(Sender: TObject; AData: TStream; ABinding: TIdSocketHandle);
  public
    { Public-Deklarationen }
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    procedure Next;
    function Data(n:integer): OLEVariant;
    function Address: string;
    function BundleID: DWord;
    function TypeTag: string;
    function TimeStamp: DWord;
    function Data1: OLEVariant;
    function Data2: OLEVariant;
    function Data3: OLEVariant;
    function Data4: OLEVariant;
    function MsgCnt: integer;
    function MsgParamCnt: integer;
    function getVersion: string;
  published
    { Published-Deklarationen }
//    property Canvas;
    property Visible;
    property Font;
//    property Brush;
    property Color;
    property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property Indicator: integer read FIndicator write FIndicator;
    property ListeningPort: integer read FListeningPort write SetListeningPort;
    property OnOscInput: TNotifyEvent read FOnOscInput write FOnOscInput;
    property Version : string read getVersion;
  end;

procedure Register;

implementation

{ $R OSC.RES}

function ReadOSCString(AData:TStream):string;
var
  s: String;
  a: array[1..4] of byte;
  i,j : integer;
  endfound : boolean;
begin
  endfound := false;
  s := '';
  repeat
    j := AData.Read(a,4);
    for i := 1 to j do
      if a[i] > 0
        then s := s+chr(a[i])
        else endfound := true;
    until endfound or (j < 4);
  result := s;
end;

function ReadOSCLong(AData:TStream):int64;
var
  a,b: array[1..8] of byte;
  i,j : integer;
begin
  j := AData.Read(a,8);
  for i := 1 to j do b[9-i] := a[i];
  result := int64(b);
end;

function ReadOSCInt(AData:TStream):integer;
var
  a,b: array[1..4] of byte;
  i,j : integer;
begin
  for i := 1 to 4 do b[i] := 0;
  j := AData.Read(a,4);
  for i := 1 to j do b[5-i] := a[i];
  result := integer(b);
end;

function ReadOSCSingle(AData:TStream):single;
var
  a,b: array[1..4] of byte;
  i,j : integer;
begin
  j := AData.Read(a,4);
  for i := 1 to j do b[5-i] := a[i];
  result := single(b);
end;


//---------- TOSC

constructor TOSC.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FUDPServer := tIdUDPServer.Create(Self);
  OSCMsgList := TObjectList.Create;
  Width := 50; Height := 33;
  FProgress := 2*BevelWidth;
  Indicator := 5;
  FBundleID := 0;
  Color := clSkyBlue;
  BevelInner := bvLowered;
  Flisteningport := 7000;
  Caption := '';
  Repaint;
end;

destructor TOSC.Destroy;
begin
  Close;
  FUDPServer.Free;
  OscMsgList.Clear;
  OScMsgList.Free;
  inherited Destroy;
end;

procedure TOSC.Paint;
var
//  X, Y,
  W, H,n: Integer;
  capt: array[0..50] of char;
  R: TRect;
//  s : string;
//  c : char;
  k: integer;
const
  cPropeller : array [0..3] of char = ('*','+','#','5');
begin
  Canvas.Font := Font;
  Canvas.Brush := Brush;
  with Canvas do
  begin
//    Pen := FPen;
//    Brush := FBrush;
    inherited;
//    X := Pen.Width div 2;
//    Y := X;
    W := Width - Pen.Width + 1;
    H := Height - Pen.Width + 1;
//    Rectangle(X, Y, X + W, Y + H);
    k := 2*BevelWidth;
    R := Rect(k,k,W-1-k,H-1-k);

    n := OSCMsgList.Count+1;
    if n > width-k then n := width-k;
    if FProgress >= Width-n-k
      then FProgress := k;

{    if n > 0
    then begin
      inc (FPropeller);
      c := cPropeller[FPropeller mod 3];
      if n > 30
      then s := c+'...'
      else if n > 10                1
      then s := c+'.. '
      else if n > 2
      then s := c+'.  '
      else s := c+'   '
    end else s := '    ';
}
    formatbuf(capt,16,'OSC Port %5d '#0,14,[FListeningport]);
    DrawText(Canvas.Handle, Capt, -1, R, DT_Center or DT_NOPREFIX or
      DT_WORDBREAK or DrawTextBiDiModeFlagsReadingOnly);

    //Indicator
    Canvas.Pen.Color := clRed;
    Canvas.Brush.Color := clRed;
    Rectangle(FProgress,h-k-Indicator,Fprogress+n,h-k);
  end;
end;

procedure TOSC.SetListeningPort(const Value: integer);
begin
  FUDPServer.Active := False;
  FUDPServer.Bindings.Clear;
  FListeningPort := value;
  FUDPServer.DefaultPort := value;
  Open
//  FUDPServer.ThreadedEvent := true;
end;

procedure TOSC.Open;
begin
  if FUDPServer.DefaultPort > 0 then
  try
    FUDPServer.Active := True;
    FUDPServer.OnUDPRead := UDPServerUDPRead;
  except end;
  repaint;
end;

procedure TOSC.Close;
begin
  if assigned(FUDPServer)
  then begin
    FUDPServer.Active := False;
    FUDPServer.OnUDPRead := nil;
    FUDPServer.Bindings.Clear;  //essential fr reconnect
  end;
end;

procedure TOSC.Next;
begin
  if OSCMsgList.count > 0 then
  begin
//    tOscMsg(OSCMsgList.Items[0]).free;
    OSCMsgList.delete(0);
  end;
end;

procedure Register;
begin
  RegisterComponents('Beispiele', [TOSC]);
end;

//---------- TOSC property functions
function TOSC.Data(n:integer): OLEVariant;
begin
//  dec(n);
  if (OSCMsgList.count > 0) and (MsgParamCnt > n) then
  begin
    result := tOscMsg(OSCMsgList.Items[0]).varia[n];
  end else
    result := '0';
end;

function TOSC.Data1: OLEVariant;
begin
  if (OSCMsgList.count > 0) and (length(tOscMsg(OSCMsgList.Items[0]).varia) > 0) then
  begin
    result := tOscMsg(OSCMsgList.Items[0]).varia[0];
  end else
    result := '0';
end;

function TOSC.Data2: OLEVariant;
begin
  if (OSCMsgList.count > 0) and (length(tOscMsg(OSCMsgList.Items[0]).varia) > 1) then
  begin
    result := tOscMsg(OSCMsgList.Items[0]).varia[1];
  end else
    result := '0';
end;

function TOSC.Data3: OLEVariant;
begin
  if (OSCMsgList.count > 0) and (length(tOscMsg(OSCMsgList.Items[0]).varia) > 2) then
  begin
    result := tOscMsg(OSCMsgList.Items[0]).varia[2];
  end else
    result := '0';
end;

function TOSC.Data4: OLEVariant;
begin
  if (OSCMsgList.count > 0) and (length(tOscMsg(OSCMsgList.Items[0]).varia) > 3) then
  begin
    result := tOscMsg(OSCMsgList.Items[0]).varia[2];
  end else
    result := '0';
end;

function TOSC.Address: string;
begin
  if (OSCMsgList.count > 0)
    then result := tOscMsg(OSCMsgList.Items[0]).address
    else result := '';
end;

function TOSC.TypeTag: string;
begin
  if (OSCMsgList.count > 0)
    then result := tOscMsg(OSCMsgList.Items[0]).typetag
    else result := '';
end;

function TOSC.TimeStamp: DWord;
begin
  if (OSCMsgList.count > 0)
    then result := tOscMsg(OSCMsgList.Items[0]).timestamp
    else result := 0;
end;

function TOSC.BundleID: DWord;
begin
  if (OSCMsgList.count > 0)
    then result := tOscMsg(OSCMsgList.Items[0]).bundleID
    else result := 0;
end;

function TOSC.MsgCnt: integer;
begin
  result := OSCMsgList.count;
end;

function TOSC.MsgParamCnt: integer;
begin
    if (OSCMsgList.count > 0)
    then result := length(tOscMsg(OSCMsgList.Items[0]).varia)
    else result := 0;
end;

function TOSC.GetVersion : string;
begin
  result := Ver;
end;

procedure TOSC.UDPServerUDPRead(Sender: TObject; AData: TStream; ABinding: TIdSocketHandle);
var
  AOscMsg : TOscMsg;
  adr,s,ss: String;
  fbundle : boolean;
  i,j,k : integer;
  f : single;
  l : int64;
  a : byte;
//  siz: int64;
  iVar : integer;
begin
  if FUDPServer.Active and assigned(Adata) and (AData.Size > 0) and (ABinding.Handle >= 0)
  then begin

//    siz := AData.Size;

    Tag := Tag + 1;

    //OSC Nachricht zerlegen
    try
      s := ReadOSCString(AData);

      fbundle := s = '#bundle';
      if fBundle
        then begin
            l := ReadOSCLong(AData);  //Time-Tag
            k := ReadOscInt(AData);   //length
            inc(FBundleID);
            //Adress
            adr := ReadOSCString(AData);
        end else begin
          adr := s;
          l := 0;
          k := 1;
        end;

       //Create List Entry
       AOscMsg := TOscMsg.Create;
       AOscMsg.address := adr;
       AOscMsg.timestamp := l;
       if fBundle then AOscMsg.bundleID := FBundleID else AoscMsg.bundleID := 0;
       setlength(AOscMsg.varia,0);
       while k > 0
       do begin

        //Format Tag
        iVar := 0;
        s := ReadOSCString(AData);   //Format Tag
        AOscMsg.typetag := s;

        if (s>'')
        then begin
          if (s[1] = ',')
          then begin          //type tagged message
            setlength(AOscMsg.varia,length(s)-1);
            for i := 2 to length(s) do
              case char(s[i]) of
                'f' : begin
                        f := ReadOSCSingle(AData);
                         AOscMsg.varia[iVar] := f;
                        inc (iVar);
                      end;
                'i' : begin
                        j := ReadOSCInt(AData);
                        AOscMsg.varia[iVar] := j;
                        inc (iVar);
                      end;
                's' : begin
                        ss := ReadOSCString(AData);
                        AOscMsg.varia[iVar] := ss;
                        inc (iVar);
                      end
              end;
          end else begin      //untyped message
            setlength(AOscMsg.varia,1);
            ss := '';
            for i := 1 to length(s) do
              ss := ss+inttohex(ord(s[i]),2)+' ';
            //well, as we took a string, there was a 0 at the end.
            ss := ss+ '00 ';
            for i := Adata.Position to AData.Size-1 do
            begin
              j := AData.Read(a,1);
              if j > 0 then ss := ss+inttohex(a,2)+' ';
            end;
            ss := ss+ '[untyped]';
            AOscMsg.varia[iVar] := ss;
          end
        end;
        //Messgage processed, possibly do more from bundle
        //maybe limit Count;
        OSCMsgList.Add(AOscMsg);
        inc(FProgress);

        if fBundle
        then begin
          k := ReadOscInt(AData);

          //Adress
          if k > 0 then begin
            adr := ReadOSCString(AData);

             //Create List Entry
             AOscMsg := TOscMsg.Create;
             AOscMsg.address := adr;
             AOscMsg.timestamp := l;
             AOscMsg.bundleID := FBundleID;
          end
        end else
          k := 0;
      end;
    finally
  //  DataStringStream.Free;
    end;

    repaint;
    
    if Assigned(FOnOSCInput) // and (FState = misOpen) and (GetEventCount > 0)
    then
      FOnOSCInput(Self);

  end;
end;

end.
